@@ -42,6 +42,7 @@ namespace Far { namespace internal { class FeatureMask; } }
 
 namespace Far {
 
+template <typename REAL> class PrimvarRefinerReal;
 template <class MESH> class TopologyRefinerFactory;
 
 ///
@@ -117,10 +118,13 @@ public:
     struct UniformOptions {
 
         UniformOptions(int level) :
-            refinementLevel(level),
+            refinementLevel(level & 0xf),
             orderVerticesFromFacesFirst(false),
             fullTopologyInLastLevel(false) { }
 
+        /// \brief Set uniform refinement level
+        void SetRefinementLevel(int level) { refinementLevel = level & 0xf; }
+
         unsigned int refinementLevel:4,             ///< Number of refinement iterations
                      orderVerticesFromFacesFirst:1, ///< Order child vertices from faces first
                                                     ///< instead of child vertices of vertices
@@ -152,13 +156,19 @@ public:
     struct AdaptiveOptions {
 
         AdaptiveOptions(int level) :
-            isolationLevel(level),
-            secondaryLevel(15),
+            isolationLevel(level & 0xf),
+            secondaryLevel(0xf),
             useSingleCreasePatch(false),
             useInfSharpPatch(false),
             considerFVarChannels(false),
             orderVerticesFromFacesFirst(false) { }
 
+        /// \brief Set isolation level
+        void SetIsolationLevel(int level) { isolationLevel = level & 0xf; }
+
+        /// \brief Set secondary isolation level
+        void SetSecondaryLevel(int level) { secondaryLevel = level & 0xf; }
+
         unsigned int isolationLevel:4;              ///< Number of iterations applied to isolate
                                                     ///< extraordinary vertices and creases
         unsigned int secondaryLevel:4;              ///< Shallower level to stop isolation of
@@ -173,11 +183,14 @@ public:
                                                     ///< instead of child vertices of vertices
     };
 
-    /// \brief Feature Adaptive topology refinement (restricted to scheme Catmark)
+    /// \brief Feature Adaptive topology refinement
     ///
-    /// @param options   Options controlling adaptive refinement
+    /// @param options         Options controlling adaptive refinement
     ///
-    void RefineAdaptive(AdaptiveOptions options);
+    /// @param selectedFaces   Limit adaptive refinement to the specified faces
+    ///
+    void RefineAdaptive(AdaptiveOptions options,
+                        ConstIndexArray selectedFaces = ConstIndexArray());
 
     /// \brief Returns the options specified on refinement
     AdaptiveOptions GetAdaptiveOptions() const { return _adaptiveOptions; }
@@ -212,8 +225,14 @@ protected:
     friend class PatchTableBuilder;
     friend class PatchBuilder;
     friend class PtexIndices;
-    friend class PrimvarRefiner;
+    template <typename REAL>
+    friend class PrimvarRefinerReal;
 
+    //  Copy constructor exposed via the factory class:
+    TopologyRefiner(TopologyRefiner const & source);
+
+public:
+    //  Levels and Refinements available internally (avoids need for more friends)
     Vtr::internal::Level & getLevel(int l) { return *_levels[l]; }
     Vtr::internal::Level const & getLevel(int l) const { return *_levels[l]; }
 
@@ -223,11 +242,13 @@ protected:
 private:
     //  Not default constructible or copyable:
     TopologyRefiner() : _uniformOptions(0), _adaptiveOptions(0) { }
-    TopologyRefiner(TopologyRefiner const &) : _uniformOptions(0), _adaptiveOptions(0) { }
     TopologyRefiner & operator=(TopologyRefiner const &) { return *this; }
 
     void selectFeatureAdaptiveComponents(Vtr::internal::SparseSelector& selector,
-                                         internal::FeatureMask const & mask);
+                                         internal::FeatureMask const & mask,
+                                         ConstIndexArray selectedFaces);
+    void selectLinearIrregularFaces(Vtr::internal::SparseSelector& selector,
+                                    ConstIndexArray selectedFaces);
 
     void initializeInventory();
     void updateInventory(Vtr::internal::Level const & newLevel);
@@ -241,9 +262,11 @@ private:
     Sdc::SchemeType _subdivType;
     Sdc::Options    _subdivOptions;
 
-    unsigned int _isUniform : 1,
-                 _hasHoles : 1,
-                 _maxLevel : 4;
+    unsigned int _isUniform     : 1;
+    unsigned int _hasHoles      : 1;
+    unsigned int _hasIrregFaces : 1;
+    unsigned int _regFaceSize   : 3;
+    unsigned int _maxLevel      : 4;
 
     //  Options assigned on refinement:
     UniformOptions  _uniformOptions;
@@ -256,7 +279,9 @@ private:
     int _totalFaceVertices;
     int _maxValence;
 
-    //  There is some redundancy here -- to be reduced later
+    //  Note the base level may be shared with another instance
+    bool _baseLevelOwned;
+
     std::vector<Vtr::internal::Level *>      _levels;
     std::vector<Vtr::internal::Refinement *> _refinements;
 
